# :coding: utf-8

import os
import errno
import logging

import numpy as np
import imageio
import scipy.misc


def load_image(image_path, image_size=None):
    """Return 3-D Numpy array from image *path*.

    *image_size* can be specified to resize the image.

    """
    logging.debug("Load image from path: {!r}".format(image_path))

    matrix = imageio.imread(image_path).astype(np.float)

    # If the image is monochrome, make it into an RGB image.
    if not (len(matrix.shape) == 3 and matrix.shape[2] == 3):
        matrix = np.dstack((matrix, matrix, matrix))

    if image_size is not None:
        matrix = scipy.misc.imresize(matrix, image_size)

    return matrix


def save_image(image_matrix, path):
    """Save *image_matrix* to *path*."""
    image = np.clip(image_matrix, 0, 255).astype(np.uint8)
    imageio.imwrite(path, image)


def fetch_images(path):
    """Return list of image paths from *path*."""
    if not os.path.isdir(path) or not os.access(path, os.R_OK):
        raise OSError("The image folder '{}' is incorrect".format(path))

    images = []

    for image in os.listdir(path):
        images.append(os.path.join(path, image))

    return images


def ensure_directory_access(path):
    """Ensure directory exists at *path* and is accessible."""
    try:
        os.makedirs(path)
    except OSError as error:
        if error.errno != errno.EEXIST:
            raise

        if not os.path.isdir(path):
            raise



# :coding: utf-8

import os
import logging
import operator
import functools
import time
import uuid

import tensorflow as tf
import numpy as np

import stylish.vgg
import stylish.transform
import stylish.filesystem


BATCH_SIZE = 4
CONTENT_WEIGHT = 7.5e0
STYLE_WEIGHT = 1e2
TV_WEIGHT = 2e2
LEARNING_RATE = 1e-3
EPOCHS_NUMBER = 2


def extract_model(
    style_target, content_targets, layers, mean_pixel, model_path
):
    """Train and return style generator model path.

    *style_target* should be the path to an image from which the style features
    should be extracted.

    *content_targets* should be the list of image paths from which the content
    features should be extracted.

    *layers* should be an array of layers :func:`extracted
    <stylish.vgg.extract_data>` from the Vgg19 model file.

    *mean_pixel* should be an array of three mean pixel values for the Red,
    Green and Blue channels :func:`extracted <stylish.vgg.extract_data>` from
    the Vgg19 model file.

    *model_path* should be the path where the trained model should be saved.

    """
    logging.info("Train style generator model.")

    # Pre-compute style feature map.
    style_features = compute_style_features(style_target, layers, mean_pixel)

    # Initiate a default graph.
    graph = tf.Graph()

    # Initiate the batch shape.
    batch_shape = (BATCH_SIZE, 256, 256, 3)

    with graph.as_default(), tf.Session() as session:
        content_image = tf.placeholder(
            tf.float32, shape=batch_shape, name="content_image"
        )
        normalized_image = content_image - mean_pixel
        loss_network = stylish.vgg.extract_network(normalized_image, layers)

        content_features = loss_network[stylish.vgg.CONTENT_LAYER]
        predictions = stylish.transform.network(content_image/255.0)

        loss = compute_loss_ratio(
            predictions, batch_shape, style_features, content_features,
            layers, mean_pixel
        )

        optimizer = tf.train.AdamOptimizer(LEARNING_RATE)
        training_op = optimizer.minimize(loss)

        # Start training.

        logging.info("Start training.")

        session.run(tf.global_variables_initializer())
        saver = tf.train.Saver()

        train_size = len(content_targets)

        for epoch in range(EPOCHS_NUMBER):
            logging.debug("Start epoch #{}.".format(epoch))

            start_time = time.time()

            for iteration in range(train_size // BATCH_SIZE):
                logging.debug("Start processing batch #{}.".format(iteration))
                _start_time = time.time()

                x_batch = get_next_batch(
                    iteration, content_targets, BATCH_SIZE, batch_shape
                )

                session.run(training_op, feed_dict={"X_content": x_batch})

                _end_time = time.time()
                logging.debug(
                    "Batch #{} processed [time: {}].".format(
                        iteration, _end_time - _start_time
                    )
                )

            end_time = time.time()
            logging.debug(
                "Epoch #{} processed [time: {}].".format(
                    epoch, end_time - start_time
                )
            )

            model_name = "style_model_{}".format(uuid.uuid4())
            return saver.save(session, os.path.join(model_path, model_name))


def compute_style_features(style_target, layers, mean_pixel):
    """Return computed style features map from *image_matrix*.

    The style feature map will be used to penalize the predicted image when it
    deviates from the style (colors, textures, common patterns, etc.).

    *style_target* should be the path to an image from which the style features
    should be extracted.

    *layers* should be an array of layers :func:`extracted
    <stylish.vgg.extract_data>` from the Vgg19 model file.

    *mean_pixel* should be an array of three mean pixel values for the Red,
    Green and Blue channels :func:`extracted <stylish.vgg.extract_data>` from
    the Vgg19 model file.

    """
    logging.info("Compute style features for image.")

    # Extract image matrix from image.
    image_matrix = stylish.filesystem.load_image(style_target)

    # Initiate a default graph.
    graph = tf.Graph()

    # Initiate the shape of a 4-D Tensor for a list of images.
    image_shape = (1,) + image_matrix.shape

    # Initiate the style features.
    style_features = {}

    with graph.as_default(), tf.Session() as session:
        style_image = tf.placeholder(
            tf.float32, shape=image_shape, name="style_image"
        )
        normalized_image = style_image - mean_pixel
        loss_network = stylish.vgg.extract_network(normalized_image, layers)

        # Initiate input as a list of images.
        images = np.array([image_matrix])

        for layer in stylish.vgg.STYLE_LAYERS:
            logging.debug("Start processing style layer {!r}.".format(layer))

            features = session.run(
                loss_network[layer], feed_dict={style_image: images}
            )

            logging.debug("Layer {!r} processed.".format(layer))

            features = np.reshape(features, (-1, features.shape[3]))
            gram = np.matmul(features.T, features) / features.size
            style_features[layer] = gram

    return style_features


def compute_loss_ratio(
    predictions, batch_shape, style_features, content_features,
    layers, mean_pixel
):
    """Compute loss ratio from *predictions*.

    *predictions* should be the output tensor of the
    :func:`transformation network <stylish.transform.network>`.

    *batch_shape* should be the shape of the 4-D input image list tensor.

    *style_features* should be the style features map :func:`extracted
    <compute_style_features>`.

    *content_features* should be the content features map extracted.

    *layers* should be an array of layers :func:`extracted
    <stylish.vgg.extract_data>` from the Vgg19 model file.

    *mean_pixel* should be an array of three mean pixel values for the Red,
    Green and Blue channels :func:`extracted <stylish.vgg.extract_data>` from
    the Vgg19 model file.

    """
    normalized_predictions = predictions - mean_pixel

    network = stylish.vgg.extract_network(normalized_predictions, layers)

    # Compute feature reconstruction loss from content feature map.

    logging.info("Compute feature reconstruction loss ratio.")

    content_size = _extract_tensor_size(content_features) * BATCH_SIZE
    content_loss = CONTENT_WEIGHT * (
        2 * tf.nn.l2_loss(
            network[stylish.vgg.CONTENT_LAYER] - content_features
        ) / content_size
    )

    # Compute style reconstruction loss from style features map.

    logging.info("Compute style reconstruction loss ratio.")

    style_losses = []

    for style_layer in stylish.vgg.STYLE_LAYERS:
        layer = network[style_layer]
        batch, height, width, filters = [
            dimension.value for dimension in layer.get_shape()
        ]

        size = height * width * filters
        feats = tf.reshape(layer, (batch, height * width, filters))
        feats_transposed = tf.transpose(feats, perm=[0, 2, 1])
        grams = tf.matmul(feats_transposed, feats) / size
        style_gram = style_features[style_layer]
        style_losses.append(
            2 * tf.nn.l2_loss(grams - style_gram) / style_gram.size
        )

    style_loss = (
        STYLE_WEIGHT * functools.reduce(tf.add, style_losses) / BATCH_SIZE
    )

    # Compute total variation de-noising.

    logging.info("Compute total variation loss ratio.")

    tv_y_size = _extract_tensor_size(predictions[:, 1:, :, :])
    tv_x_size = _extract_tensor_size(predictions[:, :, 1:, :])

    y_tv = tf.nn.l2_loss(
        predictions[:, 1:, :, :] - predictions[:, :batch_shape[1] - 1, :, :]
    )
    x_tv = tf.nn.l2_loss(
        predictions[:, :, 1:, :] - predictions[:, :, :batch_shape[2] - 1, :]
    )
    total_variation_loss = (
        TV_WEIGHT * 2 * (x_tv / tv_x_size + y_tv / tv_y_size) / BATCH_SIZE
    )

    return content_loss + style_loss + total_variation_loss


def get_next_batch(iteration, content_targets, batch_size, batch_shape):
    """Return Numpy array with image matrices according to *iteration* index.

    *iteration* should be an integer specifying the current portion of the
    images to return.

    *content_targets* should be the list of image paths from which the content
    features should be extracted.

    *batch_size* should be the size of the image list to return.

    *batch_shape* should be indicate the dimensions in which each image should
    be resized to.

    """
    current = iteration * batch_size
    step = current + batch_size

    x_batch = np.zeros(batch_shape, dtype=np.float32)

    # Extract and resize images from training data.
    for index, image_path in enumerate(content_targets[current:step]):
        x_batch[index] = stylish.filesystem.load_image(
            image_path, image_size=batch_shape[1:]
        ).astype(np.float32)

    return x_batch


def _extract_tensor_size(tensor):
    """Extract dimension from *tensor*."""
    return functools.reduce(
        operator.mul, (dim.value for dim in tensor.get_shape()[1:]), 1
    )



# :coding: utf-8

import sys
import argparse
import logging

import stylish.vgg
import stylish.train
import stylish.feed_forward
import stylish.filesystem


def construct_parser():
    """Return argument parser."""
    parser = argparse.ArgumentParser(
        prog="stylish",
        description="Style transfer using deep neural network.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "-v", "--verbosity",
        help="Set the logging output verbosity.",
        choices=["debug", "info", "warning", "error"],
        default="info"
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        dest="subcommands"
    )

    train_parser = subparsers.add_parser(
        "train", help="Train a style generator model.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    train_parser.add_argument(
        "--vgg19-path",
        help="Path to Vgg19 pre-trained model in the MatConvNet data format.",
        metavar="PATH",
        required=True
    )

    train_parser.add_argument(
        "--style-target",
        help="Path to image from which the style features will be extracted.",
        metavar="PATH",
        required=True
    )

    train_parser.add_argument(
        "--content-target",
        help=(
            "Path to a folder containing images from which the content "
            "features will be extracted."
        ),
        metavar="PATH",
        nargs="+",
        required=True
    )

    train_parser.add_argument(
        "--output-model",
        help="Path to folder in which the trained model will be saved.",
        metavar="PATH",
        required=True
    )

    apply_parser = subparsers.add_parser(
        "apply", help="Apply a style generator model to an image.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    apply_parser.add_argument(
        "--model",
        help=(
            "Path to trained style generator model which will be used to "
            "apply the style."
        ),
        metavar="PATH",
        required=True
    )

    apply_parser.add_argument(
        "--input",
        help="Path to image to transform.",
        metavar="PATH",
        required=True
    )

    apply_parser.add_argument(
        "--output",
        help="Path to folder in which the transformed image will be saved.",
        metavar="PATH",
        required=True
    )

    return parser


def _log_level(name):
    """Return log level from *name*.
    """
    if name == "debug":
        return logging.DEBUG

    elif name == "info":
        return logging.INFO

    elif name == "warning":
        return logging.WARNING

    elif name == "error":
        return logging.ERROR

    else:
        raise ValueError(
            "The logging level is incorrect: {!r}".format(name)
        )


def main(arguments=None):
    """Stylish command line interface."""
    if arguments is None:
        arguments = []

    # Process arguments.
    parser = construct_parser()
    namespace = parser.parse_args(arguments)

    # Initiate logger.
    logging.basicConfig(
        stream=sys.stderr, level=_log_level(namespace.verbosity),
        format="%(levelname)s: %(message)s"
    )

    if namespace.subcommands == "train":
        layers, mean_pixel = stylish.vgg.extract_data(namespace.vgg19_path)

        # Extract targeted images for training.
        content_targets = stylish.filesystem.fetch_images(
            namespace.content_target
        )

        # Ensure that the output path exist and is accessible.
        stylish.filesystem.ensure_directory_access(namespace.output_model)

        # Train the model and export it.
        model_path = stylish.train.extract_model(
            namespace.style_target, content_targets, layers, mean_pixel,
            namespace.output_model
        )

        logging.info("Model trained: {}".format(model_path))

    elif namespace.subcommands == "apply":
        # Ensure that the output path exist and is accessible.
        stylish.filesystem.ensure_directory_access(namespace.output)

        stylish.feed_forward.transform_image(
            namespace.model, namespace.input, namespace.output
        )



# :coding: utf-8

import sys

import stylish.command_line


def main():
    """Execute main command line interface passing command line arguments."""
    stylish.command_line.main(sys.argv[1:])


if __name__ == "__main__":
    main()



# :coding: utf-8

"""The Vgg19 model pre-trained for image classification is used as a loss
network in order to define perceptual loss functions that measure perceptual
differences in content and style between images.

The loss network remains fixed during the training process.

.. seealso::

    Johnson et al. (2016). Perceptual losses for real-time style transfer and
    superresolution. :ref:`CoRR, abs/1603.08155
    <https://arxiv.org/abs/1603.08155>`.

.. seealso::

    Simonyan et al. (2014). Very Deep Convolutional Networks for
    Large-Scale Image Recognition. :ref:`CoRR, abs/1409.1556
    <https://arxiv.org/abs/1409.1556>`.

    And the corresponding :ref:`Vgg19 pre-trained model
    <http://www.robots.ox.ac.uk/~vgg/research/very_deep/>` in the MatConvNet
    data format.

"""

import logging

import tensorflow as tf
import numpy as np
import scipy.io


#: List of layers which constitute the loss network.
_LAYERS = (
    "conv1_1", "relu1_1", "conv1_2", "relu1_2", "pool1",

    "conv2_1", "relu2_1", "conv2_2", "relu2_2", "pool2",

    "conv3_1", "relu3_1", "conv3_2", "relu3_2", "conv3_3", "relu3_3", "conv3_4",
    "relu3_4", "pool3",

    "conv4_1", "relu4_1", "conv4_2", "relu4_2", "conv4_3", "relu4_3", "conv4_4",
    "relu4_4", "pool4",

    "conv5_1", "relu5_1", "conv5_2", "relu5_2", "conv5_3", "relu5_3", "conv5_4",
    "relu5_4"
)

#: List of layers used to extract the style features.
STYLE_LAYERS = ("relu1_1", "relu2_1", "relu3_1", "relu4_1", "relu5_1")

#: Layer used to extract the content features.
CONTENT_LAYER = "relu4_2"


def extract_data(path):
    """Return layers and mean pixel arrays from Vgg19 model *data_path*.

    *path* should be the path to the Vgg19 pre-trained model in the
    MatConvNet data format.

    .. seealso::

        http://www.vlfeat.org/matconvnet/pretrained/

    Raise :exc:`ValueError` if the model loaded is incorrect.

    """
    logging.info("Extract data from: {!r}".format(path))

    data = scipy.io.loadmat(path)
    if not all(i in data for i in ("layers", "classes", "normalization")):
        raise ValueError("You VGG19 model file is incorrect.")

    mean = data["normalization"][0][0][0]
    mean_pixel = np.mean(mean, axis=(0, 1))
    layers = data["layers"][0]
    return layers, mean_pixel


def extract_network(input_image, layers):
    """Extract and apply the network from Vgg19 *layers*.

    A mapping with all nodes from the network is returned. The network will be
    applied to the current Tensorflow graph.

    Example::

        >>> g = tf.Graph()
        >>> with g.as_default(), tf.Session() as session:
        ...     ...
        ...     extract_network(input_image, layers)

        {
            "conv1_1": Tensor("conv2d"),
            "relu1_1": Tensor("relu"),
            "conv1_2": Tensor("conv2d"),
            "relu1_2": Tensor("relu"),
            ...
        }

    *input_image* should be a 3-D Tensor representing an image of undefined size
    with 3 channels (Red, Green and Blue). It will be the input of the network.

    *layers* should be an array of layers :func:`extracted <extract_data>` from
    the Vgg19 model file.

    """
    logging.debug("Extract network from Vgg19 pre-trained model.")

    mapping = {}

    current = input_image

    for index, name in enumerate(_LAYERS):
        layer_type = name[:4]

        if layer_type == "conv":
            weights, bias = layers[index][0][0][0][0]

            # The weight array extracted from the MatConvNet data format
            # must be transposed to match the tensorflow format.
            #
            # from: [width, height, in_channels, out_channels]
            # to: [height, width, in_channels, out_channels]
            weights = np.transpose(weights, (1, 0, 2, 3))
            bias = bias.reshape(-1)

            layer = tf.nn.conv2d(
                current, tf.constant(weights),
                strides=(1, 1, 1, 1),
                padding="SAME"
            )
            current = tf.nn.bias_add(layer, bias)

        elif layer_type == "relu":
            current = tf.nn.relu(current)

        elif layer_type == "pool":
            current = tf.nn.max_pool(
                current,
                ksize=(1, 2, 2, 1),
                strides=(1, 2, 2, 1),
                padding="SAME"
            )

        mapping[name] = current

    return mapping



# :coding: utf-8

import os
import time
import logging

import tensorflow as tf
import numpy as np

import stylish.transform
import stylish.filesystem


def transform_image(model_file, input_image, output_path):
    """Transform input image with style generator model and return result.

    *model_file* should be the path to a Tensorflow checkpoint model file.

    *input_image* should be the path to an image to transform.

    *output_path* should be the folder where the output image should be saved.

    """
    logging.info("Apply style generator model.")

    # Extract image matrix from input image.
    image_matrix = stylish.filesystem.load_image(input_image)

    # Compute output image path.
    _input_image, _ = os.path.splitext(input_image)
    output_image = os.path.join(
        output_path, "{}.jpg".format(os.path.basename(_input_image))
    )

    # Initiate a default graph.
    graph = tf.Graph()

    # Initiate the batch shape.
    batch_shape = (1,) + image_matrix.shape

    soft_config = tf.ConfigProto(allow_soft_placement=True)
    soft_config.gpu_options.allow_growth = True

    with graph.as_default(), tf.Session(config=soft_config) as session:
        images = tf.placeholder(tf.float32, shape=batch_shape, name="images")
        transform = stylish.transform.network(images)

        logging.debug("Restore style generation model.")

        saver = tf.train.Saver()
        saver.restore(session, model_file)

        logging.debug("Start transformation.")

        start_time = time.time()

        predictions = session.run(
            transform, feed_dict={images: np.array([image_matrix])}
        )
        stylish.filesystem.save_image(predictions[0], output_image)

        end_time = time.time()
        logging.info(
            "Image transformed: {} [time={}]".format(
                output_image, end_time - start_time
            )
        )

        return output_image



# :coding: utf-8

"""Stylish documentation build configuration file."""

import os
import sys
import re

# -- General ------------------------------------------------------------------

# Inject source onto path so that autodoc can find it by default, but in such a
# way as to allow overriding location.
sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "source"))
)

# Extensions.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinxcontrib.autoprogram",
    "lowdown"
]


# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"Stylish"
copyright = u"2018, Jeremy Retailleau"

# Version
with open(
    os.path.join(
        os.path.dirname(__file__), "..", "source",
        "stylish", "_version.py"
    )
) as _version_file:
    _version = re.match(
        r".*__version__ = \"(.*?)\"", _version_file.read(), re.DOTALL
    ).group(1)

version = _version
release = _version

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_template"]

# A list of prefixes to ignore for module listings.
modindex_common_prefix = [
    "stylish."
]


# -- HTML output --------------------------------------------------------------

html_theme = "sphinx_rtd_theme"

html_static_path = ["_static"]

# If True, copy source rst files to output for reference.
html_copy_source = True


# -- Autodoc ------------------------------------------------------------------

autodoc_default_flags = ["members", "undoc-members", "inherited-members"]
autodoc_member_order = "bysource"


def autodoc_skip(app, what, name, obj, skip, options):
    """Don"t skip __init__ method for autodoc."""
    if name == "__init__":
        return False

    return skip


# -- Intersphinx --------------------------------------------------------------

intersphinx_mapping = {
    "python": ("http://docs.python.org/", None)
}


# -- Setup --------------------------------------------------------------------

def setup(app):
    app.connect("autodoc-skip-member", autodoc_skip)



# :coding: utf-8

import os
import re

from setuptools import setup, find_packages


ROOT_PATH = os.path.dirname(os.path.realpath(__file__))
RESOURCE_PATH = os.path.join(ROOT_PATH, "resource")
SOURCE_PATH = os.path.join(ROOT_PATH, "source")
README_PATH = os.path.join(ROOT_PATH, "README.rst")

PACKAGE_NAME = "stylish"

# Read version from source.
with open(
    os.path.join(SOURCE_PATH, PACKAGE_NAME, "_version.py")
) as _version_file:
    VERSION = re.match(
        r".*__version__ = \"(.*?)\"", _version_file.read(), re.DOTALL
    ).group(1)


# Compute dependencies.
INSTALL_REQUIRES = [
    "tensorflow >= 1, < 2",
    "numpy >= 1.11.2, < 2",
    "scipy >= 0.1.0, < 2",
    "imageio >= 2.3.0, < 3"
]
DOC_REQUIRES = [
    "sphinx >= 1.2.2, < 2",
    "sphinx_rtd_theme >= 0.1.6, < 1",
    "lowdown >= 0.1.0, < 2",

    # Restricted as 0.1.3 causes failed builds.
    # https://bitbucket.org/birkenfeld/sphinx-contrib/issues/168
    "sphinxcontrib-autoprogram >= 0.1.2, !=0.1.3, < 1"

]
TEST_REQUIRES = [
    "pytest-runner >= 2.7, < 3",
    "pytest >= 3, < 4",
    "pytest-mock >= 0.11, < 1",
    "pytest-catchlog >= 1, < 2",
    "pytest-xdist >= 1.1, < 2",
    "pytest-cov >= 2, < 3"
]

setup(
    name="stylish",
    version=VERSION,
    description="Style transfer using deep neural network.",
    long_description=open(README_PATH).read(),
    url="http://github.com/buddly27/stylish",
    keywords=["tensorflow", "style", "transfer", "CNN"],
    author="Jeremy Retailleau",
    packages=find_packages(SOURCE_PATH),
    package_dir={
        "": "source"
    },
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    tests_require=TEST_REQUIRES,
    extras_require={
        "doc": DOC_REQUIRES,
        "test": TEST_REQUIRES,
        "dev": DOC_REQUIRES + TEST_REQUIRES
    },
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "stylish = stylish.__main__:main"
        ]
    },
)



# :coding: utf-8

"""The image transformation network is a deep residual convolutional neural
network parameterized by weights.

The network body consists of five residual blocks. All non-residual
convolutional layers are followed by an instance normalization and ReLU
non-linearities with the exception of the output layer, which instead uses a
scaled "tanh" to ensure that the output image has pixels in the range [0, 255].
Other than the first and last layers which use 9 × 9 kernels, all convolutional
layers use 3 × 3 kernels.

.. seealso::

    Johnson et al. (2016). Perceptual losses for real-time style transfer and
    superresolution. :ref:`CoRR, abs/1603.08155
    <https://arxiv.org/abs/1603.08155>`.

.. seealso::

    Ulyanov et al. (2017). Instance Normalization: The Missing Ingredient for
    Fast Stylization. :ref:`CoRR, abs/1607.08022
    <https://arxiv.org/abs/1607.08022>`.

"""

import tensorflow as tf


def network(input_images):
    """Apply the image transformation network.

    The last tensor of the graph will be returned. The network will be applied
    to the current Tensorflow graph.

    Example::

        >>> g = tf.Graph()
        >>> with g.as_default(), tf.Session() as session:
        ...     ...
        ...     network(input_images)

        Tensor("tanh"),

    *input_images* should be a 4-D Tensor representing a batch list of images.
    It will be the input of the network.

    """
    conv1 = add_convolution_layer(
        input_images, channels=32, kernel_size=9, strides=1, activation=True
    )
    conv2 = add_convolution_layer(
        conv1, channels=64, kernel_size=3, strides=2, activation=True
    )
    conv3 = add_convolution_layer(
        conv2, channels=128, kernel_size=3, strides=2, activation=True
    )

    resid1 = add_residual_block(conv3, kernel_size=3)
    resid2 = add_residual_block(resid1, kernel_size=3)
    resid3 = add_residual_block(resid2, kernel_size=3)
    resid4 = add_residual_block(resid3, kernel_size=3)
    resid5 = add_residual_block(resid4, kernel_size=3)

    deconv1 = add_deconvolution_layer(
        resid5, channels=64, kernel_size=3, strides=2, activation=True
    )
    deconv2 = add_deconvolution_layer(
        deconv1, channels=32, kernel_size=3, strides=2, activation=True
    )

    conv4 = add_convolution_layer(
        deconv2, channels=3, kernel_size=9, strides=1
    )
    output = tf.nn.tanh(conv4) * 150 + 255.0/2
    return output


def add_residual_block(input_tensor, kernel_size):
    """Apply a residual block to the network.

    *input_tensor* will be the input of the block.

    *kernel_size* should be the width and height of the convolution matrix used
    within the block.

    """
    conv1 = add_convolution_layer(input_tensor, 128, kernel_size, 1)
    relu = tf.nn.relu(conv1)
    conv2 = add_convolution_layer(relu, 128, kernel_size, 1)
    return input_tensor + conv2


def add_convolution_layer(
    input_tensor, channels, kernel_size, strides, activation=None
):
    """Apply a 2-D convolution layer to the network.

    *input_tensor* will be the input of the layer.

    *channels* should be the number of channels used as an input.

    *kernel_size* should be the width and height of the convolution matrix used
    within the block.

    *strides* should indicate the stride of the sliding window for each
    dimension of *input_tensor*.

    *activation* should indicate whether a 'relu' tensor should be added after
    the convolution layer.

    """
    _channels = input_tensor.shape[-1].value
    weights_shape = [kernel_size, kernel_size, _channels, channels]
    weights_init = tf.Variable(
        tf.truncated_normal(weights_shape, stddev=0.1, seed=1), dtype=tf.float32
    )

    strides_shape = [1, strides, strides, 1]
    tensor = tf.nn.conv2d(
        input_tensor, weights_init, strides_shape, padding="SAME"
    )

    tensor = add_instance_normalization(tensor)
    if activation is not None:
        tensor = tf.nn.relu(tensor)

    return tensor


def add_deconvolution_layer(
    input_tensor, channels, kernel_size, strides, activation=None
):
    """Apply a transposed 2-D convolution layer to the network.

    *input_tensor* will be the input of the layer.

    *channels* should be the number of channels used as an output.

    *kernel_size* should be the width and height of the convolution matrix used
    within the block.

    *strides* should indicate the stride of the sliding window for each
    dimension of *input_tensor*.

    *activation* should indicate whether a 'relu' tensor should be added after
    the convolution layer.

    """
    _channels = input_tensor.shape[-1].value
    weights_shape = [kernel_size, kernel_size, channels, _channels]
    weights_init = tf.Variable(
        tf.truncated_normal(weights_shape, stddev=0.1, seed=1), dtype=tf.float32
    )

    batch_size, rows, columns, _ = [
        dimension.value for dimension in input_tensor.get_shape()
    ]
    strides_shape = [1, strides, strides, 1]
    new_rows, new_columns = int(rows * strides), int(columns * strides)
    new_shape = [batch_size, new_rows, new_columns, channels]
    tf_shape = tf.stack(new_shape)

    tensor = tf.nn.conv2d_transpose(
        input_tensor, weights_init, tf_shape, strides_shape, padding="SAME"
    )

    tensor = add_instance_normalization(tensor)
    if activation is not None:
        tensor = tf.nn.relu(tensor)

    return tensor


def add_instance_normalization(input_tensor):
    """Apply an instance normalization to the network.

    *input_tensor* will be the input of the layer.

    .. seealso::

        Ulyanov et al. (2017). Instance Normalization: The Missing Ingredient
        for Fast Stylization. :ref:`CoRR, abs/1607.08022
        <https://arxiv.org/abs/1607.08022>`.

    """
    _channels = input_tensor.shape[-1].value
    variable_shape = [_channels]
    mu, sigma_sq = tf.nn.moments(input_tensor, [1, 2], keep_dims=True)

    shift = tf.Variable(tf.zeros(variable_shape))
    scale = tf.Variable(tf.ones(variable_shape))
    epsilon = 1e-3
    normalized = (input_tensor - mu) / (sigma_sq + epsilon)**.5

    return scale * normalized + shift

